#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import codecs
import itertools
import os
import sys

from setuptools import find_namespace_packages, setup

try:
    import versioneer
except ImportError:
    # we have a versioneer.py file living in the same directory as this file, but
    # if we're using pep 517/518 to build from pyproject.toml its not going to find it
    # https://github.com/python-versioneer/python-versioneer/issues/193#issue-408237852
    # make this work by adding this directory to the python path
    sys.path.append(os.path.dirname(os.path.realpath(__file__)))
    import versioneer


def read_requirements(filename):
    base = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(base, filename), "rb", "utf-8") as f:
        lineiter = (line.strip() for line in f)
        return [line for line in lineiter if line and not line.startswith("#")]


_dev = read_requirements("requirements/dev.txt")
_nvt = read_requirements("requirements/nvtabular.txt")
requirements = {
    "base": read_requirements("requirements/base.txt"),
    "tensorflow": read_requirements("requirements/tensorflow.txt"),
    "pytorch": read_requirements("requirements/pytorch.txt"),
    "lightfm": read_requirements("requirements/lightfm.txt"),
    "implicit": read_requirements("requirements/implicit.txt"),
    "xgboost": read_requirements("requirements/xgboost.txt"),
    "nvtabular": _nvt,
    "dev": _dev,
}
dev_requirements = {
    "tensorflow-dev": requirements["tensorflow"] + _dev + _nvt,
    "pytorch-dev": requirements["pytorch"] + _dev + _nvt,
    "implicit-dev": requirements["implicit"] + _dev + _nvt,
    "lightfm-dev": requirements["lightfm"] + _dev + _nvt,
    "xgboost-dev": requirements["xgboost"] + _dev + _nvt,
}

setup(
    name="merlin-models",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    packages=find_namespace_packages(include=["merlin*"]),
    url="https://github.com/NVIDIA-Merlin/models",
    author="NVIDIA Corporation",
    license="Apache 2.0",
    long_description=open("README.md", encoding="utf8").read(),
    long_description_content_type="text/markdown",
    install_requires=requirements["base"],
    test_suite="tests",
    extras_require={
        **requirements,
        **dev_requirements,
        "all": list(itertools.chain(*list(requirements.values()))),
    },
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Software Development :: Libraries",
        "Topic :: Scientific/Engineering",
    ],
)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import tensorflow as tf

from merlin.models.tf.metrics.evaluation import ItemCoverageAt, NoveltyAt, PopularityBiasAt


@pytest.fixture
def popularity_metrics_test_data():
    labels = tf.convert_to_tensor([[1], [3], [7], [6]], tf.int64)
    predictions = tf.convert_to_tensor(
        [[1, 9, 8, 7, 6], [1, 4, 3, 2, 5], [5, 9, 8, 7, 6]], tf.int64
    )
    item_freq_probs = [0.1, 0.05, 0.2, 0.05, 0.1, 0.1, 0.15, 0.05, 0.1, 0.1]
    return labels, predictions, item_freq_probs


@pytest.mark.parametrize("is_prob_distribution", [True, False])
def test_novelty_at_k(popularity_metrics_test_data, is_prob_distribution):
    labels, predictions, item_freq_probs = popularity_metrics_test_data
    metric1 = NoveltyAt(
        item_freq_probs=item_freq_probs, is_prob_distribution=is_prob_distribution, k=3
    )

    metric1.update_state(predicted_ids=predictions)
    result1 = metric1.result()
    tf.debugging.assert_near(result1, 2.5336342)


@pytest.mark.parametrize("is_prob_distribution", [True, False])
def test_popularity_at_k(popularity_metrics_test_data, is_prob_distribution):
    labels, predictions, item_freq_probs = popularity_metrics_test_data

    metric1 = PopularityBiasAt(
        item_freq_probs=item_freq_probs, is_prob_distribution=is_prob_distribution, k=3
    )

    metric1.update_state(predicted_ids=predictions)
    result1 = metric1.result()
    tf.debugging.assert_near(result1, 0.0833333)


def test_item_coverage_at_k(popularity_metrics_test_data):
    labels, predictions, item_freq_probs = popularity_metrics_test_data

    metric1 = ItemCoverageAt(num_unique_items=11, k=4)

    metric1.update_state(predicted_ids=predictions)
    result1 = metric1.result()
    tf.debugging.assert_near(result1, 0.727272)



import abc
import copy
from typing import Dict, List, Optional, Sequence, Union, overload

import tensorflow as tf

from merlin.models.config.schema import SchemaMixin
from merlin.models.tf.blocks.core.base import Block, BlockType, right_shift_layer
from merlin.models.tf.typing import TabularData, TensorOrTabularData
from merlin.models.tf.utils import tf_utils
from merlin.models.utils import schema_utils
from merlin.models.utils.doc_utils import docstring_parameter
from merlin.models.utils.registry import Registry, RegistryMixin
from merlin.schema import Schema, Tags

tabular_aggregation_registry: Registry = Registry.class_registry("tf.tabular_aggregations")


class TabularAggregation(
    SchemaMixin, tf.keras.layers.Layer, RegistryMixin["TabularAggregation"], abc.ABC
):
    registry = tabular_aggregation_registry

    """Aggregation of `TabularData` that outputs a single `Tensor`"""

    def call(self, inputs: TabularData, **kwargs) -> tf.Tensor:
        raise NotImplementedError()

    def _expand_non_sequential_features(self, inputs: TabularData) -> TabularData:
        inputs_sizes = {k: v.shape for k, v in inputs.items()}
        seq_features_shapes, sequence_length = self._get_seq_features_shapes(inputs_sizes)

        if len(seq_features_shapes) > 0:
            non_seq_features = set(inputs.keys()).difference(set(seq_features_shapes.keys()))
            for fname in non_seq_features:
                # Including the 2nd dim and repeating for the sequence length
                inputs[fname] = tf.tile(tf.expand_dims(inputs[fname], 1), (1, sequence_length, 1))

        return inputs

    def _get_seq_features_shapes(self, inputs_sizes: Dict[str, tf.TensorShape]):
        seq_features_shapes = dict()
        for fname, fshape in inputs_sizes.items():
            # Saves the shapes of sequential features
            if len(fshape) >= 3:
                seq_features_shapes[fname] = tuple(fshape[:2])

        sequence_length = 0
        if len(seq_features_shapes) > 0:
            if len(set(seq_features_shapes.values())) > 1:
                raise ValueError(
                    "All sequential features must share the same shape in the first two dims "
                    "(batch_size, seq_length): {}".format(seq_features_shapes)
                )

            sequence_length = list(seq_features_shapes.values())[0][1]

        return seq_features_shapes, sequence_length

    def _check_concat_shapes(self, inputs: TabularData):
        input_sizes = {k: v.shape for k, v in inputs.items()}
        if len(set([tuple(v[:-1]) for v in input_sizes.values()])) > 1:
            raise Exception(
                "All features dimensions except the last one must match: {}".format(input_sizes)
            )

    def _get_agg_output_size(self, input_size, agg_dim, axis=-1):
        batch_size = tf_utils.calculate_batch_size_from_input_shapes(input_size)
        seq_features_shapes, sequence_length = self._get_seq_features_shapes(input_size)

        if len(seq_features_shapes) > 0:
            return batch_size, sequence_length, agg_dim

        return tf.TensorShape((batch_size, agg_dim))

    def get_values(self, inputs: TabularData) -> List[tf.Tensor]:
        values = []
        for value in inputs.values():
            if type(value) is dict:
                values.extend(self.get_values(value))  # type: ignore
            else:
                values.append(value)

        return values


TabularAggregationType = Union[str, TabularAggregation]

TABULAR_MODULE_PARAMS_DOCSTRING = """
    pre: Union[str, TabularTransformation, List[str], List[TabularTransformation]], optional
        Transformations to apply on the inputs when the module is called (so **before** `call`).
    post: Union[str, TabularTransformation, List[str], List[TabularTransformation]], optional
        Transformations to apply on the inputs after the module is called (so **after** `call`).
    aggregation: Union[str, TabularAggregation], optional
        Aggregation to apply after processing the `call`-method to output a single Tensor.

        Next to providing a class that extends TabularAggregation, it's also possible to provide
        the name that the class is registered in the `tabular_aggregation_registry`. Out of the box
        this contains: "concat", "stack", "element-wise-sum" &
        "element-wise-sum-item-multi".
    schema: Optional[DatasetSchema]
        DatasetSchema containing the columns used in this block.
    name: Optional[str]
        Name of the layer.
"""


@docstring_parameter(tabular_module_parameters=TABULAR_MODULE_PARAMS_DOCSTRING)
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class TabularBlock(Block):
    """Layer that's specialized for tabular-data by integrating many often used operations.

    Note, when extending this class, typically you want to overwrite the `compute_call_output_shape`
    method instead of the normal `compute_output_shape`. This because a Block can contain pre- and
    post-processing and the output-shapes are handled automatically in `compute_output_shape`. The
    output of `compute_call_output_shape` should be the shape that's outputted by the `call`-method.

    Parameters
    ----------
    {tabular_module_parameters}
    """

    def __init__(
        self,
        pre: Optional[BlockType] = None,
        post: Optional[BlockType] = None,
        aggregation: Optional[TabularAggregationType] = None,
        schema: Optional[Schema] = None,
        name: Optional[str] = None,
        is_input: bool = False,
        **kwargs,
    ):
        super().__init__(name=name, **kwargs)
        self.input_size = None
        self.set_pre(pre)
        self.set_post(post)
        self.set_aggregation(aggregation)
        self._is_input = is_input

        if schema:
            self.set_schema(schema)

    @property
    def is_input(self) -> bool:
        return self._is_input

    @classmethod
    def from_schema(
        cls, schema: Schema, tags=None, allow_none=True, **kwargs
    ) -> Optional["TabularBlock"]:
        """Instantiate a TabularLayer instance from a DatasetSchema.

        Parameters
        ----------
        schema
        tags
        kwargs

        Returns
        -------
        Optional[TabularModule]
        """
        schema_copy = copy.copy(schema)
        if tags:
            schema_copy = schema_copy.select_by_tag(tags)
            if not schema_copy.column_names and not allow_none:
                raise ValueError(f"No features with tags: {tags} found")

        if not schema_copy.column_names:
            return None

        return cls.from_features(schema_copy.column_names, schema=schema_copy, **kwargs)

    @classmethod
    @docstring_parameter(tabular_module_parameters=TABULAR_MODULE_PARAMS_DOCSTRING, extra_padding=4)
    def from_features(
        cls,
        features: List[str],
        pre: Optional[BlockType] = None,
        post: Optional[BlockType] = None,
        aggregation: Optional[TabularAggregationType] = None,
        name=None,
        **kwargs,
    ) -> "TabularBlock":
        """
        Initializes a TabularLayer instance where the contents of features will be filtered out

        Parameters
        ----------
        features: List[str]
            A list of feature-names that will be used as the first pre-processing op to filter out
            all other features not in this list.
        {tabular_module_parameters}

        Returns
        -------
        TabularModule
        """
        pre = [Filter(features), pre] if pre else Filter(features)  # type: ignore

        return cls(pre=pre, post=post, aggregation=aggregation, name=name, **kwargs)

    def pre_call(
        self, inputs: TabularData, transformations: Optional[BlockType] = None
    ) -> TabularData:
        """Method that's typically called before the forward method for pre-processing.

        Parameters
        ----------
        inputs: TabularData
             input-data, typically the output of the forward method.
        transformations: TabularTransformationsType, optional

        Returns
        -------
        TabularData
        """
        return self._maybe_apply_transformations(
            inputs, transformations=transformations or self.pre
        )

    def call(self, inputs: TabularData, **kwargs) -> TabularData:
        return inputs

    def post_call(
        self,
        inputs: TabularData,
        transformations: Optional[BlockType] = None,
        merge_with: Optional[Union["TabularBlock", List["TabularBlock"]]] = None,
        aggregation: Optional[TabularAggregationType] = None,
    ) -> TensorOrTabularData:
        """Method that's typically called after the forward method for post-processing.

        Parameters
        ----------
        inputs: TabularData
            input-data, typically the output of the forward method.
        transformations: TabularTransformationType, optional
            Transformations to apply on the input data.
        merge_with: Union[TabularModule, List[TabularModule]], optional
            Other TabularModule's to call and merge the outputs with.
        aggregation: TabularAggregationType, optional
            Aggregation to aggregate the output to a single Tensor.

        Returns
        -------
        TensorOrTabularData (Tensor when aggregation is set, else TabularData)
        """
        _aggregation: Optional[TabularAggregation] = None
        if aggregation:
            _aggregation = TabularAggregation.parse(aggregation)
        _aggregation = _aggregation or getattr(self, "aggregation", None)

        outputs = inputs
        if merge_with:
            if not isinstance(merge_with, list):
                merge_with = [merge_with]
            for layer_or_tensor in merge_with:
                to_add = layer_or_tensor(inputs) if callable(layer_or_tensor) else layer_or_tensor
                outputs.update(to_add)

        outputs = self._maybe_apply_transformations(
            outputs, transformations=transformations or self.post
        )

        if _aggregation:
            if self.has_schema:
                _aggregation.set_schema(self.schema)
            return _aggregation(outputs)

        return outputs

    def _maybe_apply_transformations(
        self,
        inputs: TabularData,
        transformations: Optional[BlockType] = None,
    ) -> TabularData:
        """Apply transformations to the inputs if these are defined.

        Parameters
        ----------
        inputs
        transformations

        Returns
        -------

        """
        if transformations:
            _transformations = Block.parse(transformations)
            return _transformations(inputs)

        return inputs

    def compute_call_output_shape(self, input_shapes):
        return input_shapes

    def compute_output_shape(self, input_shapes):
        if self.pre:
            input_shapes = self.pre.compute_output_shape(input_shapes)

        output_shapes = self._check_post_output_size(self.compute_call_output_shape(input_shapes))

        return output_shapes

    def build(self, input_shapes):
        super().build(input_shapes)
        output_shapes = input_shapes
        if self.pre:
            self.pre.build(input_shapes)
            output_shapes = self.pre.compute_output_shape(input_shapes)

        output_shapes = self.compute_call_output_shape(output_shapes)

        if isinstance(output_shapes, dict):
            if self.post:
                self.post.build(output_shapes)
                output_shapes = self.post.compute_output_shape(output_shapes)
            if self.aggregation:
                if self.has_schema:
                    self.aggregation.set_schema(self.schema)

                self.aggregation.build(output_shapes)

    def get_config(self):
        config = super(TabularBlock, self).get_config()
        config = tf_utils.maybe_serialize_keras_objects(
            self, config, ["pre", "post", "aggregation"]
        )

        if self.has_schema:
            config["schema"] = schema_utils.schema_to_tensorflow_metadata_json(self.schema)

        return config

    @property
    def is_tabular(self) -> bool:
        return True

    def __add__(self, other):
        from models.tf.blocks.core.combinators import ParallelBlock

        return ParallelBlock(self, other)

    @classmethod
    def from_config(cls, config):
        config = tf_utils.maybe_deserialize_keras_objects(config, ["pre", "post", "aggregation"])
        if "schema" in config:
            config["schema"] = schema_utils.tensorflow_metadata_json_to_schema(config["schema"])

        return super().from_config(config)

    def _check_post_output_size(self, input_shapes):
        output_shapes = input_shapes

        if isinstance(output_shapes, dict):
            if self.post:
                output_shapes = self.post.compute_output_shape(output_shapes)
            if self.aggregation:
                schema = getattr(self, "_schema", None)
                self.aggregation.set_schema(schema)
                output_shapes = self.aggregation.compute_output_shape(output_shapes)

        return output_shapes

    def apply_to_all(self, inputs, columns_to_filter=None):
        if columns_to_filter:
            inputs = Filter(columns_to_filter)(inputs)
        outputs = tf.nest.map_structure(self, inputs)

        return outputs

    def set_schema(self, schema=None):
        self._maybe_set_schema(self.pre, schema)
        self._maybe_set_schema(self.post, schema)
        self._maybe_set_schema(self.aggregation, schema)

        return super().set_schema(schema)

    def set_pre(self, value: Optional[BlockType]):
        self._pre = Block.parse(value) if value else None

    @property
    def pre(self) -> Optional[Block]:
        """

        Returns
        -------
        SequentialTabularTransformations, optional
        """
        return self._pre

    @property
    def post(self) -> Optional[Block]:
        """

        Returns
        -------
        SequentialTabularTransformations, optional
        """
        return self._post

    def set_post(self, value: Optional[BlockType]):
        self._post = Block.parse(value) if value else None

    @property
    def aggregation(self) -> Optional[TabularAggregation]:
        """

        Returns
        -------
        TabularAggregation, optional
        """
        return self._aggregation

    def set_aggregation(self, value: Optional[Union[str, TabularAggregation]]):
        """

        Parameters
        ----------
        value
        """
        if value:
            self._aggregation: Optional[TabularAggregation] = TabularAggregation.parse(value)
        else:
            self._aggregation = None

    def repr_ignore(self):
        return []

    def repr_extra(self):
        return []

    def repr_add(self):
        return []

    @staticmethod
    def calculate_batch_size_from_input_shapes(input_shapes):
        return tf_utils.calculate_batch_size_from_input_shapes(input_shapes)

    def __rrshift__(self, other):
        return right_shift_layer(self, other)

    def super(self):
        return super()


# This is done like this to avoid mypy crashing.
def _tabular_call(  # type: ignore
    self,
    inputs: TabularData,
    *args,
    pre: Optional[BlockType] = None,
    post: Optional[BlockType] = None,
    merge_with: Optional[Union["TabularBlock", List["TabularBlock"]]] = None,
    aggregation: Optional[TabularAggregationType] = None,
    **kwargs,
) -> TensorOrTabularData:
    """We overwrite the call method in order to be able to do pre- and post-processing.

    Parameters
    ----------
    inputs: TabularData
        Input TabularData.
    pre: TabularTransformationsType, optional
        Transformations to apply before calling the forward method. If pre is None, this method
        will check if `self.pre` is set.
    post: TabularTransformationsType, optional
        Transformations to apply after calling the forward method. If post is None, this method
        will check if `self.post` is set.
    merge_with: Union[TabularModule, List[TabularModule]]
        Other TabularModule's to call and merge the outputs with.
    aggregation: TabularAggregationType, optional
        Aggregation to aggregate the output to a single Tensor.

    Returns
    -------
    TensorOrTabularData (Tensor when aggregation is set, else TabularData)
    """
    inputs = self.pre_call(inputs, transformations=pre)

    # This will call the `call` method implemented by the super class.
    outputs = self.super().__call__(inputs, *args, **kwargs)  # type: ignore

    if isinstance(outputs, dict):
        outputs = self.post_call(
            outputs, transformations=post, merge_with=merge_with, aggregation=aggregation
        )

    return outputs


TabularBlock.__call__ = _tabular_call  # type: ignore


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class Filter(TabularBlock):
    """Transformation that filters out certain features from `TabularData`."

    Parameters
    ----------
    to_include: List[str]
        List of features to include in the result of calling the module
    pop: bool
        Boolean indicating whether to pop the features to exclude from the inputs dictionary.
    """

    @overload
    def __init__(
        self,
        inputs: Sequence[str],
        name=None,
        pop=False,
        exclude=False,
        add_to_context: bool = False,
        **kwargs,
    ):
        ...

    @overload
    def __init__(
        self,
        inputs: Union[Schema, Tags],
        name=None,
        pop=False,
        exclude=False,
        add_to_context: bool = False,
        **kwargs,
    ):
        ...

    def __init__(
        self, inputs, name=None, pop=False, exclude=False, add_to_context: bool = False, **kwargs
    ):
        if isinstance(inputs, Tags):
            self.feature_names = inputs
        else:
            self.feature_names = list(inputs.column_names) if isinstance(inputs, Schema) else inputs
        super().__init__(name=name, **kwargs)
        self.exclude = exclude
        self.pop = pop
        self.add_to_context = add_to_context

    def set_schema(self, schema=None):
        out = super().set_schema(schema)

        if isinstance(self.feature_names, Tags):
            self.feature_names = self.schema.select_by_tag(self.feature_names).column_names

        return out

    def call(self, inputs: TabularData, **kwargs) -> TabularData:
        """Filter out features from inputs.

        Parameters
        ----------
        inputs: TabularData
            Input dictionary containing features to filter.

        Returns Filtered TabularData that only contains the feature-names in `self.to_include`.
        -------

        """
        assert isinstance(inputs, dict), "Inputs needs to be a dict"

        outputs = {k: v for k, v in inputs.items() if self.check_feature(k)}
        if self.pop:
            for key in outputs.keys():
                inputs.pop(key)

        if self.add_to_context:
            self.context.tensors.update(outputs)

            return {}

        return outputs

    def compute_call_output_shape(self, input_shape):
        if self.add_to_context:
            return {}

        outputs = {k: v for k, v in input_shape.items() if self.check_feature(k)}

        return outputs

    def check_feature(self, feature_name) -> bool:
        if self.exclude:
            return feature_name not in self.feature_names

        return feature_name in self.feature_names

    def get_config(self):
        config = super().get_config()
        config["inputs"] = self.feature_names
        config["exclude"] = self.exclude
        config["pop"] = self.pop

        return config


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class AsTabular(tf.keras.layers.Layer):
    """Converts a Tensor to TabularData by converting it to a dictionary.

    Parameters
    ----------
    output_name: str
        Name that should be used as the key in the output dictionary.
    name: str
        Name of the layer.
    """

    def __init__(self, output_name: str, name=None, **kwargs):
        super().__init__(name=name, **kwargs)
        self.output_name = output_name

    def call(self, inputs, **kwargs):
        return {self.output_name: inputs}

    def compute_output_shape(self, input_shape):
        return {self.output_name: input_shape}

    def get_config(self):
        config = super(AsTabular, self).get_config()
        config["output_name"] = self.output_name

        return config

    @property
    def is_tabular(self) -> bool:
        return True



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import glob
import os

import numpy as np


def list_files(startpath):
    """
    Util function to print the nested structure of a directory
    """
    for root, dirs, files in os.walk(startpath):
        level = root.replace(startpath, "").count(os.sep)
        indent = " " * 4 * (level)
        print("{}{}/".format(indent, os.path.basename(root)))
        subindent = " " * 4 * (level + 1)
        for f in files:
            print("{}{}".format(subindent, f))


def visualize_response(batch, response, top_k, session_col="session_id"):
    """
    Util function to extract top-k encoded item-ids from logits

    Parameters
    ----------
    batch : cudf.DataFrame
        the batch of raw data sent to triton server.
    response: tritonclient.grpc.InferResult
        the response returned by grpc client.
    top_k: int
        the `top_k` top items to retrieve from predictions.
    """
    sessions = batch[session_col].drop_duplicates().values
    predictions = response.as_numpy("output")
    top_preds = np.argpartition(predictions, -top_k, axis=1)[:, -top_k:]
    for session, next_items in zip(sessions, top_preds):
        print(
            "- Top-%s predictions for session `%s`: %s\n"
            % (top_k, session, " || ".join([str(e) for e in next_items]))
        )


def fit_and_evaluate(trainer, start_time_index, end_time_index, input_dir):
    """
    Util function for time-window based fine-tuning using the T4rec Trainer class.
    Iteratively train using data of a given index and evaluate on the validation data
    of the following index.

    Parameters
    ----------
    start_time_index: int
        The start index for training, it should match the partitions of the data directory
    end_time_index: int
        The end index for training, it should match the partitions of the  data directory
    input_dir: str
        The input directory where the parquet files were saved based on partition column

    Returns
    -------
    aot_metrics: dict
        The average over time of ranking metrics.
    """
    aot_metrics = {}
    for time_index in range(start_time_index, end_time_index + 1):
        # 1. Set data
        time_index_train = time_index
        time_index_eval = time_index + 1
        train_paths = glob.glob(os.path.join(input_dir, f"{time_index_train}/train.parquet"))
        eval_paths = glob.glob(os.path.join(input_dir, f"{time_index_eval}/valid.parquet"))

        # 2. Train on train data of time_index
        print("\n***** Launch training for day %s: *****" % time_index)
        trainer.train_dataset_or_path = train_paths
        trainer.reset_lr_scheduler()
        trainer.train()

        # 3. Evaluate on valid data of time_index+1
        trainer.eval_dataset_or_path = eval_paths
        eval_metrics = trainer.evaluate(metric_key_prefix="eval")
        print("\n***** Evaluation results for day %s:*****\n" % time_index_eval)
        for key in sorted(eval_metrics.keys()):
            if "at_" in key:
                print(" %s = %s" % (key.replace("at_", "@"), str(eval_metrics[key])))
                if "AOT_" + key.replace("at_", "@") in aot_metrics:
                    aot_metrics["AOT_" + key.replace("_at_", "@")] += [eval_metrics[key]]
                else:
                    aot_metrics["AOT_" + key.replace("_at_", "@")] = [eval_metrics[key]]

        # free GPU for next day training
        trainer.wipe_memory()

    return aot_metrics



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import torch
from torch.nn.modules.loss import _WeightedLoss


class LabelSmoothCrossEntropyLoss(_WeightedLoss):
    """Constructor for cross-entropy loss with label smoothing

    Parameters:
    ----------
    smoothing: float
        The label smoothing factor. it should be between 0 and 1.
    weight: torch.Tensor
        The tensor of weights given to each class.
    reduction: str
        Specifies the reduction to apply to the output,
        possible values are `none` | `sum` | `mean`

    Adapted from https://github.com/NingAnMe/Label-Smoothing-for-CrossEntropyLoss-PyTorch
    """

    def __init__(self, weight: torch.Tensor = None, reduction: str = "mean", smoothing=0.0):
        super().__init__(weight=weight, reduction=reduction)
        self.smoothing = smoothing
        self.weight = weight
        self.reduction = reduction

    @staticmethod
    def _smooth_one_hot(targets: torch.Tensor, n_classes: int, smoothing: float = 0.0):
        assert 0 <= smoothing < 1, f"smoothing factor {smoothing} should be between 0 and 1"
        with torch.no_grad():
            targets = (
                torch.empty(size=(targets.size(0), n_classes), device=targets.device)
                .fill_(smoothing / (n_classes - 1))
                .scatter_(1, targets.data.unsqueeze(1), 1.0 - smoothing)
            )
        return targets

    def forward(self, inputs, targets):
        targets = LabelSmoothCrossEntropyLoss._smooth_one_hot(
            targets, inputs.size(-1), self.smoothing
        )
        lsm = inputs

        if self.weight is not None:
            lsm = lsm * self.weight.unsqueeze(0)

        loss = -(targets * lsm).sum(-1)

        if self.reduction == "sum":
            loss = loss.sum()
        elif self.reduction == "mean":
            loss = loss.mean()
        elif self.reduction == "none":
            loss = loss
        else:
            raise ValueError(
                f"{self.reduction} is not supported, please choose one of the following values"
                " [`sum`, `none`, `mean`]"
            )
        return loss



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional

import tensorflow as tf

from merlin.models.tf.blocks.sampling.base import EmbeddingWithMetadata, ItemSampler
from merlin.models.tf.typing import TabularData


class InBatchSampler(ItemSampler):
    """Provides in-batch sampling [1]_ for two-tower item retrieval
    models. The implementation is very simple, as it
    just returns the current item embeddings and metadata, but it is necessary to have
    `InBatchSampler` under the same interface of other more advanced samplers
    (e.g. `CachedCrossBatchSampler`).
    In a nutshell, for a given (user,item) embeddings pair, the other in-batch item
    embeddings are used as negative items, rather than computing different embeddings
    exclusively for negative items.
    This is a popularity-biased sampling     as popular items are observed more often
    in training batches.
    P.s. Ignoring the false negatives (negative items equal to the positive ones) is
    managed by `ItemRetrievalScorer(..., sampling_downscore_false_negatives=True)`

    References
    ----------
    .. [1] Yi, Xinyang, et al. "Sampling-bias-corrected neural modeling for large corpus item
       recommendations." Proceedings of the 13th ACM Conference on Recommender Systems. 2019.

    Parameters
    ----------
    batch_size : int, optional
        The batch size. If not set it is inferred when the layer is built (first call())
    """

    def __init__(self, batch_size: Optional[int] = None, **kwargs):
        super().__init__(max_num_samples=batch_size, **kwargs)
        self._last_batch_items_embeddings: tf.Tensor = None  # type: ignore
        self._last_batch_items_metadata: TabularData = {}
        self.set_batch_size(batch_size)

    @property
    def batch_size(self) -> int:
        return self._batch_size

    def set_batch_size(self, value):
        self._batch_size = value
        if value is not None:
            self.set_max_num_samples(value)

    def build(self, input_shapes: TabularData) -> None:
        if self._batch_size is None:
            self.set_batch_size(input_shapes["embeddings"][0])

    def call(self, inputs: TabularData, training=True) -> EmbeddingWithMetadata:
        """Returns the item embeddings and item metadata from
        the current batch.
        The implementation is very simple, as it just returns the current
        item embeddings and metadata, but it is necessary to have
        `InBatchSampler` under the same interface of other more advanced samplers
        (e.g. `CachedCrossBatchSampler`).

        Parameters
        ----------
        inputs : TabularData
            Dict with two keys:
              "items_embeddings": Items embeddings tensor
              "items_metadata": Dict like `{"<feature name>": "<feature tensor>"}` which contains
              features that might be relevant for the sampler.
              The `InBatchSampler` does not use metadata features
              specifically, but "item_id" is required when using in combination with
              `ItemRetrievalScorer(..., sampling_downscore_false_negatives=True)`, so that
              false negatives are identified and downscored.
        training : bool, optional
            Flag indicating if on training mode, by default True

        Returns
        -------
        EmbeddingWithMetadata
            Value object with the sampled item embeddings and item metadata
        """
        self.add(inputs, training)
        items_embeddings = self.sample()
        return items_embeddings

    def add(self, inputs: TabularData, training=True) -> None:  # type: ignore
        self._check_inputs_batch_sizes(inputs)
        self._last_batch_items_embeddings = inputs["embeddings"]
        self._last_batch_items_metadata = inputs["metadata"]

    def sample(self) -> EmbeddingWithMetadata:
        return EmbeddingWithMetadata(
            self._last_batch_items_embeddings, self._last_batch_items_metadata
        )



from testbook import testbook

from tests.conftest import REPO_ROOT


@testbook(REPO_ROOT / "examples/01-Getting-started.ipynb", execute=False)
def test_func(tb):
    tb.inject(
        """
        import os
        os.environ["INPUT_DATA_DIR"] = "/raid/data/movielens"
        """
    )
    tb.execute()
    metrics = tb.ref("metrics")
    assert sorted(list(metrics.keys())) == [
        "loss",
        "rating_binary/binary_classification_task/auc",
        "rating_binary/binary_classification_task/binary_accuracy",
        "rating_binary/binary_classification_task/precision",
        "rating_binary/binary_classification_task/recall",
        "regularization_loss",
    ]



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import tensorflow as tf

import merlin.models.tf as ml


def test_inbatch_sampler():
    item_embeddings = tf.random.uniform(shape=(10, 5), dtype=tf.float32)
    item_ids = tf.random.uniform(shape=(10,), minval=1, maxval=10000, dtype=tf.int32)

    inbatch_sampler = ml.InBatchSampler()

    input_data = ml.EmbeddingWithMetadata(item_embeddings, {"item_id": item_ids})
    output_data = inbatch_sampler(input_data.__dict__)

    tf.assert_equal(input_data.embeddings, output_data.embeddings)
    for feat_name in output_data.metadata:
        tf.assert_equal(input_data.metadata[feat_name], output_data.metadata[feat_name])


def test_inbatch_sampler_no_metadata_features():
    embeddings = tf.random.uniform(shape=(10, 5), dtype=tf.float32)

    inbatch_sampler = ml.InBatchSampler()

    input_data = ml.EmbeddingWithMetadata(embeddings=embeddings, metadata={})
    output_data = inbatch_sampler(input_data.__dict__)

    tf.assert_equal(input_data.embeddings, output_data.embeddings)
    assert output_data.metadata == {}


def test_inbatch_sampler_metadata_diff_shape():
    item_embeddings = tf.random.uniform(shape=(10, 5), dtype=tf.float32)
    item_ids = tf.random.uniform(shape=(11,), minval=1, maxval=10000, dtype=tf.int32)

    inbatch_sampler = ml.InBatchSampler()

    input_data = ml.EmbeddingWithMetadata(item_embeddings, {"item_id": item_ids})

    with pytest.raises(Exception) as excinfo:
        _ = inbatch_sampler(input_data.__dict__)
    assert "The batch size (first dim) of embedding" in str(excinfo.value)


def test_popularity_sampler():
    num_classes = 100
    min_id = 2
    num_sampled = 10
    item_weights = tf.random.uniform(shape=(num_classes, 5), dtype=tf.float32)
    item_embeddings = tf.random.uniform(shape=(10, 5), dtype=tf.float32)
    item_ids = tf.random.uniform(shape=(10,), minval=1, maxval=num_classes, dtype=tf.int32)

    popularity_sampler = ml.PopularityBasedSampler(
        max_num_samples=num_sampled, max_id=num_classes - 1, min_id=min_id
    )

    input_data = ml.EmbeddingWithMetadata(item_embeddings, {"item_id": item_ids})
    output_data = popularity_sampler(input_data.__dict__, item_weights)

    assert len(tf.unique_with_counts(output_data.metadata["item_id"])[0]) == num_sampled
    tf.assert_equal(
        tf.nn.embedding_lookup(item_weights, output_data.metadata["item_id"]),
        output_data.embeddings,
    )
    tf.assert_equal(tf.reduce_all(output_data.metadata["item_id"] >= min_id), True)


def test_popularity_sampler_no_item_id():
    num_sampled = 10
    num_classes = 10000
    item_weights = tf.random.uniform(shape=(num_classes, 5), dtype=tf.float32)
    popularity_sampler = ml.PopularityBasedSampler(max_num_samples=num_sampled, max_id=num_classes)

    item_embeddings = tf.random.uniform(shape=(2, 5), dtype=tf.float32)
    input_data_0 = ml.EmbeddingWithMetadata(item_embeddings, metadata={})

    with pytest.raises(AssertionError) as excinfo:
        _ = popularity_sampler(input_data_0.__dict__, item_weights)
    assert "The 'item_id' metadata feature is required by PopularityBasedSampler" in str(
        excinfo.value
    )


def test_popularity_sampler_weights_dim_diff_max_num_samples():
    num_sampled = 10
    num_classes = 10000
    item_weights = tf.random.uniform(shape=(50, 5), dtype=tf.float32)
    item_ids = tf.random.uniform(shape=(10,), minval=1, maxval=num_classes, dtype=tf.int32)

    popularity_sampler = ml.PopularityBasedSampler(max_num_samples=num_sampled, max_id=num_classes)
    item_embeddings = tf.random.uniform(shape=(2, 5), dtype=tf.float32)
    input_data_0 = ml.EmbeddingWithMetadata(item_embeddings, {"item_id": item_ids})

    with pytest.raises(Exception) as excinfo:
        _ = popularity_sampler(input_data_0.__dict__, item_weights)
    assert "The first dimension of the items embeddings (50)" in str(excinfo.value)


@pytest.mark.parametrize("ignore_last_batch_on_sample", [True, False])
def test_cached_batches_sampler_add_sample_single_batch(ignore_last_batch_on_sample):
    batch_size = 10
    cached_batches_sampler = ml.CachedCrossBatchSampler(
        capacity=batch_size * 2,
        ignore_last_batch_on_sample=ignore_last_batch_on_sample,
    )

    item_embeddings = tf.random.uniform(shape=(batch_size, 5), dtype=tf.float32)
    item_ids = tf.random.uniform(shape=(batch_size,), minval=1, maxval=10000, dtype=tf.int32)
    item_image_embeddings = tf.random.uniform(shape=(batch_size, 8), dtype=tf.float32)

    input_data = ml.EmbeddingWithMetadata(
        item_embeddings, {"item_id": item_ids, "image": item_image_embeddings}
    )
    output_data = cached_batches_sampler(input_data.__dict__)

    if ignore_last_batch_on_sample:
        tf.assert_equal(tf.shape(output_data.embeddings)[0], 0)
        tf.assert_equal(tf.shape(output_data.metadata["item_id"])[0], 0)
        tf.assert_equal(tf.shape(output_data.metadata["image"])[0], 0)
    else:
        tf.assert_equal(output_data.embeddings, item_embeddings)
        tf.assert_equal(output_data.metadata["item_id"], item_ids)
        tf.assert_equal(output_data.metadata["image"], item_image_embeddings)


def test_cached_batches_sampler_add_sample_only_item_embedding():
    batch_size = 10
    cached_batches_sampler = ml.CachedCrossBatchSampler(
        capacity=batch_size * 2,
        ignore_last_batch_on_sample=False,
    )

    item_embeddings = tf.random.uniform(shape=(batch_size, 5), dtype=tf.float32)

    input_data = ml.EmbeddingWithMetadata(item_embeddings, {})
    output_data = cached_batches_sampler(input_data.__dict__)

    tf.assert_equal(output_data.embeddings, item_embeddings)
    assert output_data.metadata == {}


def test_cached_batches_sampler_add_diff_item_emb_metadata():
    batch_size = 10
    cached_batches_sampler = ml.CachedCrossBatchSampler(
        capacity=batch_size * 2,
        ignore_last_batch_on_sample=False,
    )

    item_embeddings = tf.random.uniform(shape=(batch_size, 5), dtype=tf.float32)
    item_ids = tf.random.uniform(shape=(batch_size + 1,), minval=1, maxval=10000, dtype=tf.int32)

    input_data = ml.EmbeddingWithMetadata(item_embeddings, {"item_id": item_ids})

    with pytest.raises(Exception) as excinfo:
        _ = cached_batches_sampler(input_data.__dict__)
    assert "The batch size (first dim) of embedding" in str(excinfo.value)


def test_cached_batches_sampler_trying_add_sample_before_built():
    cached_batches_sampler = ml.CachedCrossBatchSampler(
        capacity=20,
        ignore_last_batch_on_sample=False,
    )

    with pytest.raises(Exception) as excinfo:
        input_data = ml.EmbeddingWithMetadata(embeddings=None, metadata={})
        cached_batches_sampler.add(input_data)
    assert "The CachedCrossBatchSampler layer was not built yet." in str(excinfo.value)

    with pytest.raises(Exception) as excinfo:
        input_data = ml.EmbeddingWithMetadata(embeddings=None, metadata={})
        _ = cached_batches_sampler.sample()
    assert "The CachedCrossBatchSampler layer was not built yet." in str(excinfo.value)


@pytest.mark.parametrize("ignore_last_batch_on_sample", [True, False])
def test_cached_batches_sampler_multiple_batches(ignore_last_batch_on_sample):
    batch_size = 10
    num_batches_to_cache = 2
    num_expected_samples = batch_size * num_batches_to_cache

    cached_batches_sampler = ml.CachedCrossBatchSampler(
        capacity=batch_size * num_batches_to_cache,
        ignore_last_batch_on_sample=ignore_last_batch_on_sample,
    )
    batches_items_embeddings_list = []
    batches_items_ids_list = []
    for step in range(1, 4):
        # Making last batch incomplete (lower than the batch size)
        num_rows_current_batch = 10 if step < 3 else 7
        item_embeddings = tf.random.uniform(shape=(num_rows_current_batch, 5), dtype=tf.float32)
        batches_items_embeddings_list.append(item_embeddings)
        item_ids = tf.random.uniform(
            shape=(num_rows_current_batch,), minval=1, maxval=10000, dtype=tf.int32
        )
        batches_items_ids_list.append(item_ids)

        input_data = ml.EmbeddingWithMetadata(item_embeddings, {"item_id": item_ids})
        output_data = cached_batches_sampler(input_data.__dict__)
        expected_sampled_items = min(
            batch_size * (step - 1 if ignore_last_batch_on_sample else step), num_expected_samples
        )

        # Checks if the number of sampled items matches the expected
        tf.assert_equal(tf.shape(output_data.embeddings)[0], expected_sampled_items)
        for feat_name in output_data.metadata:
            tf.assert_equal(tf.shape(output_data.metadata[feat_name])[0], expected_sampled_items)

    expected_sampled_item_embeddings = tf.concat(batches_items_embeddings_list, axis=0)
    expected_sampled_item_ids = tf.concat(batches_items_ids_list, axis=0)

    if ignore_last_batch_on_sample:
        expected_sampled_item_embeddings = expected_sampled_item_embeddings[
            :-num_rows_current_batch
        ]
        expected_sampled_item_ids = expected_sampled_item_ids[:-num_rows_current_batch]
    else:
        # Checks if the sampled items correspond to the 2nd and 3rd batches
        expected_sampled_item_embeddings = expected_sampled_item_embeddings[num_rows_current_batch:]
        expected_sampled_item_ids = expected_sampled_item_ids[num_rows_current_batch:]

    tf.assert_equal(output_data.embeddings, expected_sampled_item_embeddings)
    tf.assert_equal(output_data.metadata["item_id"], expected_sampled_item_ids)


@pytest.mark.parametrize("ignore_last_batch_on_sample", [True, False])
def test_cached_batches_sampler_max_num_samples(ignore_last_batch_on_sample):
    batch_size = 10
    num_batches_to_cache = 2

    cached_batches_sampler = ml.CachedCrossBatchSampler(
        capacity=batch_size * num_batches_to_cache,
        ignore_last_batch_on_sample=ignore_last_batch_on_sample,
    )
    assert cached_batches_sampler.max_num_samples == batch_size * num_batches_to_cache


def test_cached_uniform_sampler_adds_or_updates_items():
    queue_capacity = 15
    uniform_sampler = ml.CachedUniformSampler(
        capacity=queue_capacity, ignore_last_batch_on_sample=False
    )

    # Adding 4 new items ids but two of them with repeated item ids (should add only
    # the 2 unique ones)
    item_ids0 = tf.constant([0, 0, 1, 1], dtype=tf.int32)
    item_embeddings0 = tf.random.uniform(shape=(4, 5), dtype=tf.float32)
    input_data_0 = ml.EmbeddingWithMetadata(item_embeddings0, {"item_id": item_ids0})
    output_data_0 = uniform_sampler(input_data_0.__dict__)
    assert tuple(output_data_0.embeddings.shape) == (2, 5)
    tf.assert_equal(output_data_0.metadata["item_id"], [0, 1])
    tf.assert_equal(
        output_data_0.embeddings,
        tf.stack([item_embeddings0[0], item_embeddings0[2]], axis=0),
    )

    # Updating two existing items (0,1) and adding 8 new items ids (2-9)
    item_ids1 = tf.range(0, 10)
    item_embeddings1 = tf.random.uniform(shape=(10, 5), dtype=tf.float32)
    input_data_1 = ml.EmbeddingWithMetadata(item_embeddings1, {"item_id": item_ids1})
    output_data_1 = uniform_sampler(input_data_1.__dict__)
    assert tuple(output_data_1.embeddings.shape) == (10, 5)
    tf.assert_equal(output_data_1.metadata["item_id"], item_ids1)
    tf.assert_equal(output_data_1.embeddings, item_embeddings1)

    # Updating existing items ids (7,8,9) and adding 4 new items ids (10-12)
    item_ids2 = tf.range(7, 13)
    item_embeddings2 = tf.random.uniform(shape=(6, 5), dtype=tf.float32)
    input_data_2 = ml.EmbeddingWithMetadata(item_embeddings2, {"item_id": item_ids2})
    output_data_2 = uniform_sampler(input_data_2.__dict__)
    assert tuple(output_data_2.embeddings.shape) == (13, 5)
    tf.assert_equal(
        output_data_2.metadata["item_id"], tf.concat([item_ids1[:7], item_ids2], axis=0)
    )
    tf.assert_equal(
        output_data_2.embeddings, tf.concat([item_embeddings1[:7], item_embeddings2], axis=0)
    )

    # Updating existing items ids (0,1,2)
    item_ids3 = tf.range(0, 3)
    item_embeddings3 = tf.random.uniform(shape=(3, 5), dtype=tf.float32)
    input_data_3 = ml.EmbeddingWithMetadata(item_embeddings3, {"item_id": item_ids3})
    output_data_3 = uniform_sampler(input_data_3.__dict__)
    assert tuple(output_data_3.embeddings.shape) == (13, 5)
    tf.assert_equal(
        output_data_3.metadata["item_id"],
        tf.concat([item_ids3[0:3], item_ids1[3:7], item_ids2], axis=0),
    )
    tf.assert_equal(
        output_data_3.embeddings,
        tf.concat([item_embeddings3[0:3], item_embeddings1[3:7], item_embeddings2], axis=0),
    )

    # Adding four new items (13,14,15,16). As adding those items will exceed the queue capacity,
    # the first 2 items added to the queue will be removed to keep the queue size = 15
    item_ids4 = tf.range(13, 17)
    item_embeddings4 = tf.random.uniform(shape=(4, 5), dtype=tf.float32)
    input_data_4 = ml.EmbeddingWithMetadata(item_embeddings4, {"item_id": item_ids4})
    output_data_4 = uniform_sampler(input_data_4.__dict__)
    assert tuple(output_data_4.embeddings.shape) == (queue_capacity, 5)
    tf.assert_equal(
        output_data_4.metadata["item_id"],
        tf.concat([item_ids3[2:3], item_ids1[3:7], item_ids2, item_ids4], axis=0),
    )
    tf.assert_equal(
        output_data_4.embeddings,
        tf.concat(
            [item_embeddings3[2:3], item_embeddings1[3:7], item_embeddings2, item_embeddings4],
            axis=0,
        ),
    )


@pytest.mark.parametrize("ignore_last_batch_on_sample", [True, False])
def test_cached_uniform_sampler_expected_number_examples(ignore_last_batch_on_sample):
    queue_capacity = 100
    uniform_sampler = ml.CachedUniformSampler(
        capacity=queue_capacity, ignore_last_batch_on_sample=ignore_last_batch_on_sample
    )

    for step in range(1, 4):
        batch_size = 10
        # Ensures item ids are unique
        item_ids = tf.range(step * batch_size, (step * batch_size) + 10)
        item_embeddings = tf.random.uniform(shape=(batch_size, 5), dtype=tf.float32)
        input_data = ml.EmbeddingWithMetadata(item_embeddings, {"item_id": item_ids})
        output_data = uniform_sampler(input_data.__dict__)

        expected_samples = batch_size * (step - 1 if ignore_last_batch_on_sample else step)
        assert tuple(output_data.embeddings.shape) == (expected_samples, 5)
        assert tuple(output_data.metadata["item_id"].shape) == (expected_samples,)


def test_cached_uniform_sampler_no_item_id():
    uniform_sampler = ml.CachedUniformSampler(
        capacity=10,
    )
    item_embeddings = tf.random.uniform(shape=(4, 5), dtype=tf.float32)
    input_data_0 = ml.EmbeddingWithMetadata(item_embeddings, metadata={})

    with pytest.raises(AssertionError) as excinfo:
        _ = uniform_sampler(input_data_0.__dict__)
    assert "The 'item_id' metadata feature is required by UniformSampler" in str(excinfo.value)


def test_cached_uniform_sampler_trying_add_sample_before_built():
    cached_batches_sampler = ml.CachedUniformSampler(
        capacity=10,
    )

    with pytest.raises(Exception) as excinfo:
        input_data = ml.EmbeddingWithMetadata(embeddings=None, metadata={})
        cached_batches_sampler.add(input_data.__dict__)
    assert "The CachedUniformSampler layer was not built yet." in str(excinfo.value)

    with pytest.raises(Exception) as excinfo:
        input_data = ml.EmbeddingWithMetadata(embeddings=None, metadata={})
        _ = cached_batches_sampler.sample()
    assert "The CachedUniformSampler layer was not built yet." in str(excinfo.value)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import tensorflow as tf

import merlin.models.tf as ml


@pytest.mark.parametrize(
    "loss",
    [
        # Pairwise losses
        "bpr",
        ml.losses.BPRLoss(),
        "bpr-max",
        ml.losses.BPRmaxLoss(reg_lambda=1.0),
        "top1",
        ml.losses.TOP1Loss(),
        "top1_v2",
        ml.losses.TOP1v2Loss(),
        "top1-max",
        ml.losses.TOP1maxLoss(),
        "logistic",
        ml.losses.LogisticLoss(),
        "hinge",
        ml.losses.HingeLoss(),
        # Listwise losses
        "sparse_categorical_crossentropy",
        "categorical_crossentropy",
        # Pointwise losses
        "mse",
        "binary_crossentropy",
    ],
)
def test_losses(loss):
    batch_size = 100
    num_samples = 20
    predictions = tf.random.uniform(shape=(batch_size, num_samples), dtype=tf.float32)
    positives = tf.ones(shape=(batch_size, 1), dtype=tf.float32)
    negatives = tf.zeros(shape=(batch_size, num_samples - 1), dtype=tf.float32)
    targets = tf.concat([positives, negatives], axis=1)

    if loss == "sparse_categorical_crossentropy":
        targets = tf.argmax(targets, axis=1)

    loss = ml.losses.loss_registry.parse(loss)
    loss_output = loss(targets, predictions)
    assert len(tf.shape(loss_output)) == 0
    assert loss_output > 0


def test_pairwise_loss_no_reduction():
    batch_size = 100
    num_samples = 20
    predictions = tf.random.uniform(shape=(batch_size, num_samples), dtype=tf.float32)
    positives = tf.ones(shape=(batch_size, 1), dtype=tf.float32)
    negatives = tf.zeros(shape=(batch_size, num_samples - 1), dtype=tf.float32)
    targets = tf.concat([positives, negatives], axis=1)

    bpr = ml.losses.BPRLoss(reduction=tf.keras.losses.Reduction.NONE)
    loss = bpr(targets, predictions)
    tf.assert_equal(tf.shape(loss), (batch_size, num_samples - 1))
    assert tf.reduce_mean(loss) > 0


def test_pairwise_loss_with_sample_weights():
    batch_size = 100
    num_samples = 20
    predictions = tf.random.uniform(shape=(batch_size, num_samples), dtype=tf.float32)
    positives = tf.ones(shape=(batch_size, 1), dtype=tf.float32)
    negatives = tf.zeros(shape=(batch_size, num_samples - 1), dtype=tf.float32)
    targets = tf.concat([positives, negatives], axis=1)
    sample_weights = tf.range(1, 101, dtype=tf.float32)

    bpr_max = ml.losses.BPRLoss(reduction=tf.keras.losses.Reduction.NONE)
    loss = bpr_max(targets, predictions)
    loss_with_sampled_weights = bpr_max(targets, predictions, sample_weights)

    tf.assert_equal(tf.shape(loss), (batch_size, num_samples - 1))
    tf.assert_equal(tf.shape(loss_with_sampled_weights), (batch_size, num_samples - 1))

    tf.assert_equal(loss * tf.expand_dims(sample_weights, -1), loss_with_sampled_weights)

    assert (tf.reduce_mean(loss) * tf.reduce_mean(sample_weights)).numpy() == pytest.approx(
        tf.reduce_mean(loss_with_sampled_weights).numpy(), 0.05
    )


def test_pairwise_loss_multiple_positive():
    batch_size = 100
    num_samples = 20
    predictions = tf.random.uniform(shape=(batch_size, num_samples), dtype=tf.float32)
    positives = tf.ones(shape=(batch_size, 2), dtype=tf.float32)
    negatives = tf.zeros(shape=(batch_size, num_samples - 2), dtype=tf.float32)
    targets = tf.concat([positives, negatives], axis=1)

    bpr = ml.losses.BPRLoss()

    with pytest.raises(Exception) as excinfo:
        _ = bpr(targets, predictions)
    assert "The batch contains more examples with more than one positive item" in str(excinfo.value)


def test_pairwise_loss_some_rows_with_no_positive():
    batch_size = 30
    num_samples = 20
    predictions = tf.random.uniform(shape=(batch_size, num_samples), dtype=tf.float32)

    n_rows_without_positive = batch_size - num_samples
    targets = tf.eye(num_rows=batch_size, num_columns=num_samples)

    bpr = ml.losses.BPRLoss(reduction=tf.keras.losses.Reduction.NONE)
    loss = bpr(targets, predictions)

    tf.assert_equal(tf.shape(loss), (batch_size, num_samples - 1))
    tf.assert_greater(tf.reduce_min(loss[:-n_rows_without_positive]), 0.0)
    tf.assert_equal(tf.reduce_sum(loss[-n_rows_without_positive:]), 0.0)


def test_pairwise_loss_with_false_negatives():
    batch_size = 30
    num_samples = 20
    predictions = tf.random.uniform(shape=(batch_size, num_samples), dtype=tf.float32)
    positives = tf.ones(shape=(batch_size, 1), dtype=tf.float32)
    negatives = tf.zeros(shape=(batch_size, num_samples - 1), dtype=tf.float32)
    targets = tf.concat([positives, negatives], axis=1)
    valid_negatives_mask = tf.cast(
        1.0 - tf.eye(num_rows=batch_size, num_columns=num_samples - 1), tf.bool
    )

    bpr = ml.losses.BPRLoss(reduction=tf.keras.losses.Reduction.NONE)
    loss = bpr(targets, predictions, valid_negatives_mask=valid_negatives_mask)

    tf.assert_equal(tf.shape(loss), (batch_size, num_samples - 1))
    # Checks if the loss of false negatives are zero
    tf.assert_equal(tf.reduce_sum(tf.boolean_mask(loss, tf.logical_not(valid_negatives_mask))), 0.0)
    # Checks if the loss of true negatives are non-zero
    tf.assert_greater(tf.reduce_min(tf.boolean_mask(loss, valid_negatives_mask)), 0.0)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from copy import copy, deepcopy
from dataclasses import dataclass
from typing import Any, Callable, Dict, Optional, Union

import tensorflow as tf
from tensorflow.python import to_dlpack
from tensorflow.python.keras import backend
from tensorflow.python.tpu.tpu_embedding_v2_utils import FeatureConfig, TableConfig

import merlin.io
from merlin.models.tf.blocks.core.base import Block, BlockType
from merlin.models.tf.blocks.core.combinators import SequentialBlock
from merlin.models.tf.blocks.core.tabular import (
    TABULAR_MODULE_PARAMS_DOCSTRING,
    Filter,
    TabularAggregationType,
    TabularBlock,
)
from merlin.models.tf.blocks.core.transformations import AsDenseFeatures, AsSparseFeatures

# pylint has issues with TF array ops, so disable checks until fixed:
# https://github.com/PyCQA/pylint/issues/3613
# pylint: disable=no-value-for-parameter, unexpected-keyword-arg
from merlin.models.tf.typing import TabularData
from merlin.models.utils import schema_utils
from merlin.models.utils.doc_utils import docstring_parameter
from merlin.schema import Schema, Tags, TagsType

EMBEDDING_FEATURES_PARAMS_DOCSTRING = """
    feature_config: Dict[str, FeatureConfig]
        This specifies what TableConfig to use for each feature. For shared embeddings, the same
        TableConfig can be used for multiple features.
    item_id: str, optional
        The name of the feature that's used for the item_id.
"""


@dataclass
class EmbeddingOptions:
    embedding_dims: Optional[Dict[str, int]] = None
    embedding_dim_default: Optional[int] = 64
    infer_embedding_sizes: bool = False
    infer_embedding_sizes_multiplier: float = 2.0
    infer_embeddings_ensure_dim_multiple_of_8: bool = False
    embeddings_initializers: Optional[
        Union[Dict[str, Callable[[Any], None]], Callable[[Any], None]]
    ] = None
    embeddings_l2_reg: float = 0.0
    combiner: Optional[str] = "mean"


@docstring_parameter(
    tabular_module_parameters=TABULAR_MODULE_PARAMS_DOCSTRING,
    embedding_features_parameters=EMBEDDING_FEATURES_PARAMS_DOCSTRING,
)
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class EmbeddingFeatures(TabularBlock):
    """Input block for embedding-lookups for categorical features.

    For multi-hot features, the embeddings will be aggregated into a single tensor using the mean.

    Parameters
    ----------
    {embedding_features_parameters}
    {tabular_module_parameters}
    """

    def __init__(
        self,
        feature_config: Dict[str, "FeatureConfig"],
        pre: Optional[BlockType] = None,
        post: Optional[BlockType] = None,
        aggregation: Optional[TabularAggregationType] = None,
        schema: Optional[Schema] = None,
        name=None,
        add_default_pre=True,
        l2_reg: Optional[float] = 0.0,
        **kwargs,
    ):
        if add_default_pre:
            embedding_pre = [Filter(list(feature_config.keys())), AsSparseFeatures()]
            pre = [embedding_pre, pre] if pre else embedding_pre  # type: ignore
        self.feature_config = feature_config
        self.l2_reg = l2_reg
        super().__init__(
            pre=pre,
            post=post,
            aggregation=aggregation,
            name=name,
            schema=schema,
            is_input=True,
            **kwargs,
        )

    @classmethod
    def from_schema(  # type: ignore
        cls,
        schema: Schema,
        embedding_options: EmbeddingOptions = EmbeddingOptions(),
        tags: Optional[TagsType] = None,
        max_sequence_length: Optional[int] = None,
        **kwargs,
    ) -> Optional["EmbeddingFeatures"]:
        """Instantiates embedding features from the schema

        Parameters
        ----------
        schema : Schema
            The features chema
        embedding_options : EmbeddingOptions, optional
            An EmbeddingOptions instance, which allows for a number of
            options for the embedding table, by default EmbeddingOptions()
        tags : Optional[TagsType], optional
            If provided, keeps only features from those tags, by default None
        max_sequence_length : Optional[int], optional
            Maximum sequence length of sparse features (if any), by default None

        Returns
        -------
        EmbeddingFeatures
            An instance of EmbeddingFeatures block, with the embedding
            layers created under-the-hood
        """
        schema_copy = copy(schema)

        if tags:
            schema_copy = schema_copy.select_by_tag(tags)

        embedding_dims = embedding_options.embedding_dims or {}
        if embedding_options.infer_embedding_sizes:
            inferred_embedding_dims = schema_utils.get_embedding_sizes_from_schema(
                schema,
                embedding_options.infer_embedding_sizes_multiplier,
                embedding_options.infer_embeddings_ensure_dim_multiple_of_8,
            )
            # Adding inferred embedding dims only for features where the embedding sizes
            # were not pre-defined
            inferred_embedding_dims = {
                k: v for k, v in inferred_embedding_dims.items() if k not in embedding_dims
            }
            embedding_dims = {**embedding_dims, **inferred_embedding_dims}

        initializer_default = tf.keras.initializers.TruncatedNormal(mean=0.0, stddev=0.05)
        embeddings_initializer = embedding_options.embeddings_initializers or initializer_default

        emb_config = {}
        cardinalities = schema_utils.categorical_cardinalities(schema)
        for key, cardinality in cardinalities.items():
            embedding_size = embedding_dims.get(key, embedding_options.embedding_dim_default)
            if isinstance(embeddings_initializer, dict):
                emb_initializer = embeddings_initializer.get(key, initializer_default)
            else:
                emb_initializer = embeddings_initializer
            emb_config[key] = (cardinality, embedding_size, emb_initializer)

        feature_config: Dict[str, FeatureConfig] = {}
        tables: Dict[str, TableConfig] = {}

        domains = schema_utils.categorical_domains(schema)
        for name, (vocab_size, dim, emb_initilizer) in emb_config.items():
            table_name = domains[name]
            table = tables.get(table_name, None)
            if not table:
                table = TableConfig(
                    vocabulary_size=vocab_size,
                    dim=dim,
                    name=table_name,
                    combiner=embedding_options.combiner,
                    initializer=emb_initilizer,
                )
                tables[table_name] = table
            feature_config[name] = FeatureConfig(table)

        if not feature_config:
            return None

        output = cls(
            feature_config,
            schema=schema_copy,
            l2_reg=embedding_options.embeddings_l2_reg,
            **kwargs,
        )

        return output

    def build(self, input_shapes):
        self.embedding_tables = {}
        tables: Dict[str, TableConfig] = {}
        for name, feature in self.feature_config.items():
            table: TableConfig = feature.table
            if table.name not in tables:
                tables[table.name] = table

        for name, table in tables.items():
            add_fn = (
                self.context.add_embedding_weight if hasattr(self, "_context") else self.add_weight
            )
            self.embedding_tables[name] = add_fn(
                name=name,
                trainable=True,
                initializer=table.initializer,
                shape=(table.vocabulary_size, table.dim),
            )
        if isinstance(input_shapes, dict):
            super().build(input_shapes)
        else:
            tf.keras.layers.Layer.build(self, input_shapes)

    def call(self, inputs: TabularData, **kwargs) -> TabularData:
        embedded_outputs = {}
        for name, val in inputs.items():
            embedded_outputs[name] = self.lookup_feature(name, val)
            if self.l2_reg > 0:
                self.add_loss(self.l2_reg * tf.reduce_sum(tf.square(embedded_outputs[name])))

        return embedded_outputs

    def compute_call_output_shape(self, input_shapes):
        batch_size = self.calculate_batch_size_from_input_shapes(input_shapes)

        output_shapes = {}
        for name, val in input_shapes.items():
            output_shapes[name] = tf.TensorShape([batch_size, self.feature_config[name].table.dim])

        return output_shapes

    def lookup_feature(self, name, val, output_sequence=False):
        dtype = backend.dtype(val)
        if dtype != "int32" and dtype != "int64":
            val = tf.cast(val, "int32")

        table: TableConfig = self.feature_config[name].table
        table_var = self.embedding_tables[table.name]
        if isinstance(val, tf.SparseTensor):
            out = tf.nn.safe_embedding_lookup_sparse(table_var, val, None, combiner=table.combiner)
        else:
            if output_sequence:
                out = tf.gather(table_var, tf.cast(val, tf.int32))
            else:
                if len(val.shape) > 1:
                    # TODO: Check if it is correct to retrieve only the 1st element
                    # of second dim for non-sequential multi-hot categ features
                    out = tf.gather(table_var, tf.cast(val, tf.int32)[:, 0])
                else:
                    out = tf.gather(table_var, tf.cast(val, tf.int32))
        if self._dtype_policy.compute_dtype != self._dtype_policy.variable_dtype:
            # Instead of casting the variable as in most layers, cast the output, as
            # this is mathematically equivalent but is faster.
            out = tf.cast(out, self._dtype_policy.compute_dtype)

        return out

    def table_config(self, feature_name: str):
        return self.feature_config[feature_name].table

    def get_embedding_table(self, table_name: Union[str, Tags], l2_normalization: bool = False):
        if isinstance(table_name, Tags):
            feature_names = self.schema.select_by_tag(table_name).column_names
            if len(feature_names) == 1:
                table_name = feature_names[0]
            elif len(feature_names) > 1:
                raise ValueError(
                    f"There is more than one feature associated to the tag {table_name}"
                )
            else:
                raise ValueError(f"Could not find a feature associated to the tag {table_name}")

        embeddings = self.embedding_tables[table_name]
        if l2_normalization:
            embeddings = tf.linalg.l2_normalize(embeddings, axis=-1)

        return embeddings

    def embedding_table_df(
        self, table_name: Union[str, Tags], l2_normalization: bool = False, gpu: bool = True
    ):
        """Retrieves a dataframe with the embedding table

        Parameters
        ----------
        table_name : Union[str, Tags]
            Tag or name of the embedding table
        l2_normalization : bool, optional
            Whether the L2-normalization should be applied to
            embeddings (common approach for Matrix Factorization
            and Retrieval models in general), by default False
        gpu : bool, optional
            Whether or not should use GPU, by default True

        Returns
        -------
        Union[pd.DataFrame, cudf.DataFrame]
            Returns a dataframe (cudf or pandas), depending on the gpu
        """
        embeddings = self.get_embedding_table(table_name, l2_normalization)
        if gpu:
            import cudf
            import cupy

            # Note: It is not possible to convert Tensorflow tensors to the cudf dataframe
            # directly using dlPack (as the example commented below) because cudf.from_dlpack()
            # expects the 2D tensor to be in Fortran order (column-major), which is not
            # supported by TF (https://github.com/rapidsai/cudf/issues/10754).
            # df = cudf.from_dlpack(to_dlpack(tf.convert_to_tensor(embeddings)))
            embeddings_cupy = cupy.fromDlpack(to_dlpack(tf.convert_to_tensor(embeddings)))
            df = cudf.DataFrame(embeddings_cupy)
            df.columns = [str(col) for col in list(df.columns)]
            df.set_index(cudf.RangeIndex(0, embeddings.shape[0]))
        else:
            import pandas as pd

            df = pd.DataFrame(embeddings.numpy())
            df.columns = [str(col) for col in list(df.columns)]
            df.set_index(pd.RangeIndex(0, embeddings.shape[0]))

        return df

    def embedding_table_dataset(
        self, table_name: Union[str, Tags], l2_normalization: bool = False, gpu=True
    ) -> merlin.io.Dataset:
        """Creates a Dataset for the embedding table

        Parameters
        ----------
        table_name : Union[str, Tags]
            Tag or name of the embedding table
        l2_normalization : bool, optional
            Whether the L2-normalization should be applied to
            embeddings (common approach for Matrix Factorization
            and Retrieval models in general), by default False
        gpu : bool, optional
            Whether or not should use GPU, by default True

        Returns
        -------
        merlin.io.Dataset
            Returns a Dataset with the embeddings
        """
        return merlin.io.Dataset(self.embedding_table_df(table_name, l2_normalization, gpu))

    def export_embedding_table(
        self,
        table_name: Union[str, Tags],
        export_path: str,
        l2_normalization: bool = False,
        gpu=True,
    ):
        """Exports the embedding table to parquet file

        Parameters
        ----------
        table_name : Union[str, Tags]
            Tag or name of the embedding table
        export_path : str
            Path for the generated parquet file
        l2_normalization : bool, optional
            Whether the L2-normalization should be applied to
            embeddings (common approach for Matrix Factorization
            and Retrieval models in general), by default False
        gpu : bool, optional
            Whether or not should use GPU, by default True
        """
        df = self.embedding_table_df(table_name, l2_normalization, gpu=gpu)
        df.to_parquet(export_path)

    def get_config(self):
        config = super().get_config()

        feature_configs = {}

        for key, val in self.feature_config.items():
            feature_config_dict = dict(name=val.name, max_sequence_length=val.max_sequence_length)

            feature_config_dict["table"] = serialize_table_config(val.table)
            feature_configs[key] = feature_config_dict

        config["feature_config"] = feature_configs

        return config

    @classmethod
    def from_config(cls, config):
        # Deserialize feature_config
        feature_configs, table_configs = {}, {}
        for key, val in config["feature_config"].items():
            feature_params = deepcopy(val)
            table_params = feature_params["table"]
            if "name" in table_configs:
                feature_params["table"] = table_configs["name"]
            else:
                table = deserialize_table_config(table_params)
                if table.name:
                    table_configs[table.name] = table
                feature_params["table"] = table
            feature_configs[key] = FeatureConfig(**feature_params)
        config["feature_config"] = feature_configs

        # Set `add_default_pre to False` since pre will be provided from the config
        config["add_default_pre"] = False

        return super().from_config(config)


@docstring_parameter(
    tabular_module_parameters=TABULAR_MODULE_PARAMS_DOCSTRING,
    embedding_features_parameters=EMBEDDING_FEATURES_PARAMS_DOCSTRING,
)
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class SequenceEmbeddingFeatures(EmbeddingFeatures):
    """Input block for embedding-lookups for categorical features. This module produces 3-D tensors,
    this is useful for sequential models like transformers.
    Parameters
    ----------
    {embedding_features_parameters}
    padding_idx: int
        The symbol to use for padding.
    {tabular_module_parameters}
    """

    def __init__(
        self,
        feature_config: Dict[str, FeatureConfig],
        max_seq_length: Optional[int] = None,
        mask_zero: bool = True,
        padding_idx: int = 0,
        pre: Optional[BlockType] = None,
        post: Optional[BlockType] = None,
        aggregation: Optional[TabularAggregationType] = None,
        schema: Optional[Schema] = None,
        name: Optional[str] = None,
        add_default_pre=True,
        **kwargs,
    ):
        if add_default_pre:
            embedding_pre = [Filter(list(feature_config.keys())), AsDenseFeatures(max_seq_length)]
            pre = [embedding_pre, pre] if pre else embedding_pre  # type: ignore

        super().__init__(
            feature_config=feature_config,
            pre=pre,
            post=post,
            aggregation=aggregation,
            name=name,
            schema=schema,
            add_default_pre=False,
            **kwargs,
        )
        self.padding_idx = padding_idx
        self.mask_zero = mask_zero

    def lookup_feature(self, name, val, **kwargs):
        return super(SequenceEmbeddingFeatures, self).lookup_feature(
            name, val, output_sequence=True
        )

    def compute_call_output_shape(self, input_shapes):
        batch_size = self.calculate_batch_size_from_input_shapes(input_shapes)
        sequence_length = input_shapes[list(self.feature_config.keys())[0]][1]

        output_shapes = {}
        for name, val in input_shapes.items():
            output_shapes[name] = tf.TensorShape(
                [batch_size, sequence_length, self.feature_config[name].table.dim]
            )

        return output_shapes

    def compute_mask(self, inputs, mask=None):
        if not self.mask_zero:
            return None
        outputs = {}
        for key, val in inputs.items():
            outputs[key] = tf.not_equal(val, self.padding_idx)

        return outputs

    def get_config(self):
        config = super().get_config()
        config["mask_zero"] = self.mask_zero
        config["padding_idx"] = self.padding_idx

        return config


def ContinuousEmbedding(
    inputs: Block,
    embedding_block: Block,
    aggregation=None,
    continuous_aggregation="concat",
    name: str = "continuous",
    **kwargs,
) -> SequentialBlock:
    continuous_embedding = Filter(Tags.CONTINUOUS, aggregation=continuous_aggregation).connect(
        embedding_block
    )

    outputs = inputs.connect_branch(
        continuous_embedding.as_tabular(name), add_rest=True, aggregation=aggregation, **kwargs
    )

    return outputs


def serialize_table_config(table_config: TableConfig) -> Dict[str, Any]:
    table = deepcopy(table_config.__dict__)
    if "initializer" in table:
        table["initializer"] = tf.keras.initializers.serialize(table["initializer"])
    if "optimizer" in table:
        table["optimizer"] = tf.keras.optimizers.serialize(table["optimizer"])

    return table


def deserialize_table_config(table_params: Dict[str, Any]) -> TableConfig:
    if "initializer" in table_params and table_params["initializer"]:
        table_params["initializer"] = tf.keras.initializers.deserialize(table_params["initializer"])
    if "optimizer" in table_params and table_params["optimizer"]:
        table_params["optimizer"] = tf.keras.optimizers.deserialize(table_params["optimizer"])
    table = TableConfig(**table_params)

    return table


def serialize_feature_config(feature_config: FeatureConfig) -> Dict[str, Any]:
    outputs = {}

    for key, val in feature_config.items():
        feature_config_dict = dict(name=val.name, max_sequence_length=val.max_sequence_length)
        feature_config_dict["table"] = serialize_table_config(feature_config_dict["table"])
        outputs[key] = feature_config_dict

    return outputs
